#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"
 
const double 	INT_I	=0,		/* inicio do intervalo de integracao */ 
	     	INT_F	=10;		/* fim do intervalo */
	int 	N	   ;		/* numero de processos, deve coincidir com o argumento de mpiexec */
const int 	NMSG	=1;		/* no de mensagens que cada no enviara, > 1 */

 
/*
	funcao que se deseja integrar,
   	nao necessariamente conhecida.
*/
double myFunc(double x) {
	usleep(100);
	return (x);
}

int isDone(int *nrecv) {
	int c;
	/* rank==0 is not counted */
	for (c=1; c<N; c++) {
		if(nrecv[c] != NMSG) return 0; 
	}
	return 1;
}

int main(int argc, char ** argv) {

	MPI_Status status;
	int rank, size;
	double time_start, time_end;	
	double delta = 0; /*(INT_F - INT_I)/(N-1)*/
	char processor_name[MPI_MAX_PROCESSOR_NAME];
	int temp;

	if( MPI_Init(&argc, &argv) < 0 ) return -1;    

	MPI_Comm_rank(MPI_COMM_WORLD,&rank);
	MPI_Comm_size(MPI_COMM_WORLD,&size);
	MPI_Get_processor_name(processor_name,&temp);

	N=size;
	delta = (INT_F - INT_I) / (N - 1);
	fflush(stdout);
	
	MPI_Barrier(MPI_COMM_WORLD);

	if(N == 1) {
		/* programa equivalente serial, so para comparar */
		
		double result=0;
		double i,precision,INC;
		
		if(argc > 1) {
			if(strcmp(argv[1],"fixo")==0){
				INC = 0.01;
			}
			 if(strcmp(argv[1],"var")==0){
        	                INC = (INT_F - INT_I)/100;
                	}
		} else {
			printf("faltou os parametros!!!");
		}


/*		const double INC = atof(argv[1]);*/
		precision= INC/1000;

		time_start = MPI_Wtime();
		fflush(stdout);

		for(i=INT_I; i+INC-precision < INT_F; i+=INC) {
			result+= INC*0.5*( myFunc(i)+myFunc(i+INC) );
		}

                time_end=MPI_Wtime();
                printf("\n%d\t %lf\t %lf",N,result,time_end-time_start);
                fflush(stdout);

	} else 
	if(rank == 0) {

		int nrecv[N];
		double result=0;
		double p;
		int i;

		time_start= MPI_Wtime();
		fflush(stdout);		
		
		for( i=0; i<N; i++) {
			nrecv[i]=0;
		}
		while(!isDone(nrecv)) {
			MPI_Recv(&p, 1, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
			nrecv[status.MPI_SOURCE]++;
			result += p;
		}
		time_end=MPI_Wtime();
		printf("\n%d\t %lf\t %lf",N,result,time_end-time_start);
		fflush(stdout);

	} else {
		double result_parcial=0;
		double i,precision,INC;
		
		if(strcmp(argv[1],"fixo")==0){
                        INC = 0.01;
                }
		if(strcmp(argv[1],"var")==0){
                        INC = delta/100;
                }


/*		const double INC = atof(argv[1]);*/

/*		double delta_linha = delta/NMSG; */
			/* delta_linha eh usado apenas para complicar um pouco,
			   se NMSG for 1, minimiza comunicacao logo melhor 
			   performance. Posso aumentar NMSG (numero de mensagens
			   enviada por casa processo) para ver o impacto na 
			   performance global. */
                precision=INC/1000; 
			result_parcial=0;
			for ( i = (rank-1)*delta; i+INC-precision < (rank)*delta; i+=INC ) {
				result_parcial += INC*0.5* ( myFunc(i)+ myFunc(i+INC-precision) );
			}
			MPI_Send(&result_parcial, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);

	}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();
	return 0;
}


